_na = 'name'
_tp = 'type'
_opt = 'optional'
_nu = 'nullable'
_cp = 'is_compound'
_srl = 'compound_serialize-mcrr'
_sch = 'compound_schema'
_def = 'default'
_def_succ = 'default_success'
_def_fail = 'default_failure'
_miss_keys = 'missing_keys'


_spbtps = (int, float, bool, str)
_empt = (None, "", (), [], {})

_mtchs = [((int,), 3), ((int, range(1,5)), 3), ((int, lambda x: x > 0), 3),
          ((str,), 'string'), ((str, ('SUCCESS', 'FAILURE')), 'SUCCESS'),
          ((str, '^[A-Za-z0-9]{2}$'), 'a9'), ((float,), 1.2), ((bool,), True)]

_f_mtchs = [(int, 3), (str, 'string'), (float, 'string'), (bool, 3),
            ((int, 2, 3), 3), ((str,), 3), ((int,), 'string'), ((float,), 'string'),
            ((bool,), 3)]



import json
from types import LambdaType
from uuid import UUID
from serialize_mcr.hp import _valid_regex, _regex_match, _err, _spsstps
from serialize_mcr.const import _spbtps, _na, _nu, _tp, _opt, _cp, _srl, \
    _sch, _empt, _def

name = "serialize_mcr"


class SerializeMCR(object):
    schema = []

    def __init__(self, data):
        self.preproc(data)
        self.proc(data)

    def __getattr__(self, item):
        return self.item

    @staticmethod
    def valid_type(_type):
        if isinstance(_type, tuple) and len(_type) == 2:
            mt = _type[0]
            st = _type[1]
            if mt not in _spbtps:
                return False
            if (mt == int and isinstance(st, tuple) is False
                    and isinstance(st, range) is False
                    and isinstance(st, LambdaType) is False):
                return False
            if mt == int and isinstance(st, tuple) is True:
                for e in st:
                    if isinstance(e, int) is False:
                        return False
            if (mt == str and isinstance(st, tuple) is False
                    and st not in _spsstps.keys()
                    and _valid_regex(st) is False):
                return False
            if mt == str and isinstance(st, tuple) is True:
                for e in st:
                    if isinstance(e, str) is False:
                        return False
            return True
        elif isinstance(_type, tuple) is True and len(_type) == 1:
            mt = _type[0]
            if mt not in _spbtps:
                return False
            return True
        else:
            return False

    @staticmethod
    def validated(_type, data):
        if isinstance(_type, tuple) is True and len(_type) == 2:
            mt = _type[0]
            st = _type[1]
            if mt == int and (isinstance(st, tuple) is True or isinstance(st, range) is True):
                return data in st
            elif mt == int and isinstance(st, LambdaType) is True:
                return st(data)
            elif mt == str and isinstance(st, tuple) is True:
                return data in st
            elif mt == str and st in _spsstps.keys():
                return _spsstps[st](data)
            else:
                return _regex_match(regex=st, _str=data)
        elif isinstance(_type, tuple) is True and len(_type) == 1:
            mt = _type[0]
            return isinstance(data, mt)
        else:
            return False

    def preproc(self, data):
        for prop in self.schema:
            _name = prop[_na]
            _type = prop[_tp] if _tp in prop else None
            _optional = prop[_opt] if _opt in prop else False
            if isinstance(_optional, bool) is False:
                raise TypeError(_err(8, _name=_opt, _type=type(_optional), _expected=bool))
            _nullable = prop[_nu] if _nu in prop else False
            if isinstance(_nullable, bool) is False:
                raise TypeError(_err(8, _name=_nu, _type=type(_nullable), _expected=bool))
            _compound = prop[_cp] if _cp in prop else False
            if isinstance(_compound, bool) is False:
                raise TypeError(_err(8, _name=_cp, _type=type(_compound), _expected=bool))

            _serializer = prop[_srl] if _srl in prop else None
            _schema = prop[_sch] if _sch in prop else None

            # _default -> 'value', 1, (), [], or {}
            _default = prop[_def] if _def in prop else None

            if _type is True and self.valid_type(_type) is False:
                raise TypeError(_err(5, _name, _type))
            if _default is True and isinstance(_default, _type) is False:
                raise TypeError(_err(7, _name, _type, _default))
            if _optional is False and _default is True:
                raise ValueError(_err(6, _name))
            if _optional is False and _name not in data:
                raise ValueError(_err(0, _name, data))
            if _name in data:
                if data[_name] in _empt:
                    if _nullable is False and _default is None and _optional is False:
                        raise ValueError(_err(1, _name))
                else:
                    validated = self.validated(_type, data[_name])
                    if _type is not None and validated is False:
                        raise ValueError(_err(4, _name, _type, data[_name]))
                    if (_compound is True and isinstance(data[_name], list) is False
                            and isinstance(data[_name], dict) is False):
                        raise TypeError(_err(2, _name))
                    if _compound is True and _serializer is False and _schema is False:
                        raise TypeError(_err(3, _name))

            else:
                if _default is None and _optional is False:
                    raise ValueError(_err(1, _name))

                elif _default is not None:
                    if type(_default) != bool and isinstance(_default, _type[0]) is False:
                        raise ValueError(_err(7, _name, _type=_type, _default=_default))
                    elif type(_default) == bool != _type[0]:
                        raise ValueError(_err(7, _name, _type=_type, _default=_default))
                    elif isinstance(_default, _type[0]) is True and self.validated(_type, _default) is False:
                        raise ValueError(_err(7, _name, _type=_type, _default=_default))

            prop[_na] = _name
            prop[_tp] = _type
            prop[_opt] = _optional
            prop[_nu] = _nullable
            prop[_cp] = _compound
            prop[_srl] = _serializer
            prop[_sch] = _schema
            prop[_def] = _default

            self.invalid_fields_check(data)

    def proc(self, data):
        for prop in self.schema:
            _name = prop[_na]
            _default = prop[_def]

            if _name in data:
                if prop[_cp]:
                    if prop[_srl]:
                        if isinstance(data[_name], list):
                            self.__dict__[_name] = [prop[_srl](o)
                                                    for o in data[_name]]
                        elif isinstance(data[_name], dict):
                            self.__dict__[_name] = prop[_srl](data[_name])
                    elif prop[_sch]:
                        _cls = SerializeMCR
                        _cls.schema = prop[_sch]
                        if isinstance(data[_name], list):
                            self.__dict__[_name] = [_cls(o)
                                                    for o in data[_name]]
                        elif isinstance(data[_name], dict):
                            self.__dict__[_name] = _cls(data[_name])
                else:
                    if isinstance(data[_name], UUID):
                        self.__dict__[_name] = str(data[_name])
                    else:
                        self.__dict__[_name] = data[_name]

            else:
                if _default:
                    self.__dict__[_name] = _default

    def invalid_fields_check(self, req_data):
        '''
        Check for invalid fields in request body
        '''
        invalid_fields = []
        property_names = [prop[_na] for prop in self.schema]
        for field in req_data:
            if field not in property_names:
                invalid_fields.append(field)

        if len(invalid_fields) > 0:
            raise KeyError(_err(e=9, _name=name, data=invalid_fields))

    def as_dict(self):
        _d = {}
        for prop in self.schema:
            _name = prop[_na]
            if prop[_cp]:
                if self.__dict__ and _name in self.__dict__.keys():
                    if isinstance(self.__dict__[_name], list):
                        _d[_name] = [cp.as_dict() for cp in self.__dict__[_name]]
                    else:
                        _d[_name] = self.__dict__[_name].as_dict()

                elif not self.__dict__:
                    raise ValueError(_err(e=10, _name=_name))

            else:
                if _name in self.__dict__:
                    _d[_name] = self.__dict__[_name]
        return _d

    def _to_str(self):
        return json.dumps(self.as_dict())

    def __str__(self) -> str:
        return self._to_str()

    def __repr__(self) -> str:
        return self._to_str()



_na = 'name'
_tp = 'type'
_opt = 'optional'
_nu = 'nullable'
_cp = 'is_compound'
_srl = 'compound_serialize-mcrr'
_sch = 'compound_schema'
_def = 'default'

_spbtps = (int, float, bool, str)

_empt = (None, "", (), [], {})



from serialize-mcr import serialize-mcr
from tests.unit import const
from tests.conftest import TestNameParameter, TestTypeParameter, TestNullableParameter, TestOptionalParameter, \
    TestDefaultParameter, TestDefaultParameterFailure, TestCompoundSchemaFunctionality, \
    TestCompoundserialize-mcrrFunctionality, TestInvalidFields, test_name_param_data_success, \
    test_name_param_data_failure, test_type_param_data_success, test_type_param_data_failure, \
    test_nullable_param_data_success, test_nullable_param_data_failure, test_optional_param_data_success, \
    test_optional_param_data_failure, test_default_param_data_success, test_default_param_data_failure, \
    test_compound_schema_data_success, test_compound_schema_data_failure, test_compound_serialize-mcrr_data_success, \
    test_compound_serialize-mcrr_data_failure, test_invalid_fields_check

import logging

class Expected(object):
    """Class contains data used for unit testing.
    Attributes:
        _status_code: the status code for http response.
        _data: the response data.
    """

    def __init__(self, status_code, data) -> None:
        self._status_code = status_code
        self._data = data

    def status_code(self):
        """A getter method"""
        return self._status_code

    def data(self):
        """A getter method"""
        return self._data


class Suite(object):
    serialize-mcrr = {}

    def __init__(self):
        self.serial_j = serialize-mcr
        self.schema = [
            {'name': 'var1', 'type': (int,), 'nullable': False, 'optional': False, 'default': None},
            {'name': 'var2', 'type': (str,), 'nullable': False, 'optional': False, 'default': None},
            {'name': 'var3', 'type': (bool,), 'nullable': True, 'optional': True, 'default': False},
            {'name': 'var4', 'type': (float,), 'nullable': True, 'optional': True, 'default': 1.25},
        ]
        self.executor = {}
        self._bind(_type=const._na,
                   serialize-mcrr=TestNameParameter,
                   suc_data=test_name_param_data_success(),
                   fail_data=test_name_param_data_failure())
        self._bind(_type=const._tp,
                   serialize-mcrr=TestTypeParameter,
                   suc_data=test_type_param_data_success(),
                   fail_data=test_type_param_data_failure())
        self._bind(_type=const._nu,
                   serialize-mcrr=TestNullableParameter,
                   suc_data=test_nullable_param_data_success(),
                   fail_data=test_nullable_param_data_failure())
        self._bind(_type=const._opt,
                   serialize-mcrr=TestOptionalParameter,
                   suc_data=test_optional_param_data_success(),
                   fail_data=test_optional_param_data_failure())
        # default is a little different
        #   two separate serialize-mcrr schemas are needed to test out a default fail/success case
        self._bind(_type=const._def_succ,
                   serialize-mcrr=TestDefaultParameter,
                   suc_data=test_default_param_data_success(),
                   fail_data=None)
        self._bind(_type=const._def_fail,
                   serialize-mcrr=TestDefaultParameterFailure,
                   suc_data=None,
                   fail_data=test_default_param_data_failure())
        self._bind(_type=const._sch,
                   serialize-mcrr=TestCompoundSchemaFunctionality,
                   suc_data=test_compound_schema_data_success(),
                   fail_data=test_compound_schema_data_failure())
        self._bind(_type=const._srl,
                   serialize-mcrr=TestCompoundserialize-mcrrFunctionality,
                   suc_data=test_compound_serialize-mcrr_data_success(),
                   fail_data=test_compound_serialize-mcrr_data_failure())

        self._bind(_type=const._miss_keys,
                   serialize-mcrr=TestInvalidFields,
                   suc_data=None,
                   fail_data=test_invalid_fields_check())

        self._execute()
        self._valid_type()
        self._validation()

    def _bind(self, _type, serialize-mcrr, suc_data=None, fail_data=None):
        self.serialize-mcrr[_type] = serialize-mcrr
        self.executor[_type] = {
            'success': suc_data,
            'failure': fail_data
        }

    def _execute(self):
        for binding in self.executor:
            if self.serialize-mcrr[binding]:
                suc_data = self.executor[binding].get('success', None)
                fail_data = self.executor[binding].get('failure', None)

                if suc_data is not None:
                    try:
                        self.serialize-mcrr[binding](data=suc_data)

                    except (ValueError, TypeError) as e:
                        raise e

                if fail_data is not None:
                    cases = fail_data.get('cases')
                    if cases is not None:
                        for case in fail_data['cases'].keys():
                            fail_test_data = fail_data['cases'][case]['data']
                            fail_test_expected = fail_data['cases'][case]['expected']

                            try:
                                self.serialize-mcrr[binding](data=fail_test_data)

                            except (ValueError, TypeError, KeyError) as e:
                                assert type(e) == type(fail_test_expected)

                            else:
                                assert False
        logging.info('Successfully passed all success and fail cases in data serialization.')

    def _valid_type(self):
        for suc_subtype in const._spbtps:
            assert self.serial_j.valid_type((suc_subtype, ))

        for fail_subtype in const._spbtps:
            assert self.serial_j.valid_type((str(fail_subtype),)) is False

        for fail_empties in const._empt:
            assert self.serial_j.valid_type((fail_empties,)) is False

        logging.info('Successfully passed all success and fail cases in the type validation.')

    def _validation(self):
        for succ_matches in const._mtchs:
            assert self.serial_j.validated(succ_matches[0], succ_matches[1])

        for fail_matches in const._f_mtchs:
            assert self.serial_j.validated(fail_matches[0], fail_matches[1]) is False

        logging.info('Successfully passed all success and fail cases in the validation method.')
